import climetlab.plotting

# List of possible projections
for p in climetlab.plotting.projections():
    print(p)

# List of possible styles
for p in climetlab.plotting.styles():
    print(p)

# List of possible backgrounds/foregrounds
for p in climetlab.plotting.layers():
    print(p)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))

import datetime
import os
import sys

top = os.path.realpath(os.path.dirname(os.path.dirname(__file__)))
sys.path.insert(0, top)


source_suffix = ".rst"
master_doc = "index"
pygments_style = "sphinx"
html_theme_options = {"logo_only": True}
html_logo = "_static/magpye.png"


# -- Project information -----------------------------------------------------

project = "magpye"

author = "ECMWF"

year = datetime.datetime.now().year
if year == 2020:
    years = "2020"
else:
    years = "2020-%s" % (year,)

copyright = "%s, ECMWF" % (years,)


# The full version, including alpha/beta/rc tags
# release = climetlab.__version__


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx_panels",
    # "sphinx.ext.intersphinx",
    # "sphinx.ext.extlinks",
    # "sphinx.ext.mathjax",
    "sphinx.ext.doctest",
    "sphinx.ext.napoleon",
    # "sphinx.ext.viewcode", # Uncomment to add links to source code
    "sphinx.ext.todo",
    "nbsphinx",
    "IPython.sphinxext.ipython_directive",
    "IPython.sphinxext.ipython_console_highlighting",
    "sphinx.ext.graphviz",
    # "sphinx-prompt",
    # "climetlab.sphinxext.sources",
    # "climetlab.sphinxext.command_output",
    # "climetlab.sphinxext.module_output",
]

# Add any paths that contain templates here, relative to this directory.
# templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "'**.ipynb_checkpoints'"]


# https://www.notion.so/Deepnote-Launch-Buttons-63c642a5e875463495ed2341e83a4b2a

nbsphinx_prolog = """
{% set docname = env.doc2path(env.docname, base=None) %}

You can run this notebook in |Binder|, in |Colab|.

.. |Binder| image:: https://mybinder.org/badge.svg
   :target: https://mybinder.org/v2/gh/ecmwf/ecmwf-geomaps/main?urlpath=lab/tree/docs/{{ docname }}
   :alt: Binder
   :class: cmlbadge


.. |Colab| image:: https://colab.research.google.com/assets/colab-badge.svg
   :target: https://colab.research.google.com/github/ecmwf/ecmwf-geomaps/blob/main/docs/{{ docname }}
   :alt: Colab
   :class: cmlbadge


"""  # noqa

todo_include_todos = True

# ipython_warning_is_error = False

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]
html_css_files = ["style.css"]

# See https://www.sphinx-doc.org/en/master/usage/extensions/graphviz.html
graphviz_output_format = "svg"




from .. import macro
from ..action import action
from ._data import Data


class Grib(Data):

    @action(
        macro.mgrib,
        {
            "grib_automatic_scaling": False,
        },
        file_name="grib_input_file_name",
        index="grib_field_position",
        grib_id="grib_id",
        grib_mode="grib_file_address_mode",
        wind_mode="grib_wind_mode",
        wind_1="grib_wind_position_1",
        wind_2="grib_wind_position_2",
    )
    def _get(self, *args, **kwargs):
        pass



from .geomap import GeoMap


def get_active_geomap():
    return GeoMapHelper.get_active()


class GeoMapHelper:

    active = None

    EXTENT_KWARGS = {
        "subpage_lower_left_latitude": lambda x, y: x < y,
        "subpage_lower_left_longitude": lambda x, y: x < y,
        "subpage_upper_right_latitude": lambda x, y: x > y,
        "subpage_upper_right_longitude": lambda x, y: x > y,
    }

    @classmethod
    def get_active(cls):
        if cls.active is None:
            cls.active = cls.default_geomap()
        return cls.active

    @classmethod
    def default_geomap(cls):
        return GeoMap(preset="background-foreground")

    @classmethod
    def _modify_bounds(cls, source):
        new_extent = source.extent
        active_extent = [
            cls.active.queue[0].kwargs.get(kwarg) for kwarg in cls.EXTENT_KWARGS
        ]
        for key, new_value in new_extent:
            active_value = active_extent.get(key, new_value)
            if cls.EXTENT_KWARGS[key](new_value, active_value):
                cls.active.queue[0].kwargs[key] = new_value



# (C) Copyright 2021- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os

import yaml

from . import config


def get(style, method_name):
    method_name = method_name.lstrip("_")
    yaml_file = os.path.join(config.PRESETS, method_name, f"{style}.yaml")
    if not os.path.exists(yaml_file):
        raise ValueError(f"no preset '{style}' defined for {method_name}()")

    with open(yaml_file, "r") as f:
        style_definition = yaml.load(f, Loader=yaml.SafeLoader)

    try:
        kwargs = style_definition["style"]
    except KeyError:
        raise TypeError("style definition requires 'style' key")

    return kwargs



#!/usr/bin/env python3
# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os

me = os.path.basename(__file__)
here = os.path.dirname(__file__)

CODE = """
import climetlab as cml
cml.plotting_options(path="{}")
{}
"""

for root, _, files in os.walk(here):
    for file in files:
        if file.endswith(".py") and file != me:
            full = os.path.join(root, file)
            with open(full) as f:
                code = f.read()
                if "plot_map" in code:
                    n = len(here) + 1
                    path = os.path.join(here, "_static", full[n:])
                    path = path.replace(".py", ".svg")
                    if not os.path.exists(os.path.dirname(path)):
                        os.makedirs(os.path.dirname(path))
                    print("PATH", path)
                    code = CODE.format(path, code)
                    exec(code, dict(__file__=full), {})




from .. import macro
from ..action import action
from ._data import Data


class NetCDF(Data):

    @action(
        macro.mnetcdf,
        {
            "netcdf_field_automatic_scaling": False,
        },
        file_name="netcdf_file_name",
        variable_name="netcdf_value_variable",
    )
    def _get(self, *args, **kwargs):
        pass



import nbformat as nbf
from Magics import macro as magics

nb = nbf.v4.new_notebook()
title = """\
# Predefined areas in **mappye**"""
import_magpye = """\
from magpye import GeoMap"""

nb.cells.append(nbf.v4.new_markdown_cell(title))
nb.cells.append(nbf.v4.new_code_cell(import_magpye))

areas = magics.predefined_areas()

for area in areas:
    title = """\
## {} """.format(
        area
    )

    code = """\
map = GeoMap(area_name="{}")
map.coastlines(land_colour="grey")
map.gridlines(line_style="dash")
map.show()""".format(
        area
    )
    nb.cells.append(nbf.v4.new_markdown_cell(title))
    nb.cells.append(nbf.v4.new_code_cell(code))

fname = "test.ipynb"

with open(fname, "w") as f:
    nbf.write(nb, f)



# (C) Copyright 2021- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from functools import wraps

from . import styles


def action(magics_macro, conditions=None, **valid_args):
    """
    Decorator for generating figure methods.

    Parameters
    ----------
    magics_macro : magpye.macro.Macro
        The Magics macro to execute for this action.
    condition : dict
        A dctionary of conditional variables which are
        required for the given action to execute on the given macro.
        Values can be sub-dictionaries, indicating that the condition is
        only required when a specific argument is passed.
    valid_args : dict
        A mapping of argument names to their Magics counterparts.
    """

    def decorator(method):
        @wraps(method)
        def wrapper(self, *args, style=None, z_index=1, **kwargs):
            for i, arg in enumerate(args):
                try:
                    kwarg = list(valid_args)[i]
                except IndexError:
                    raise TypeError(
                        f"{method.__name__}() takes a maximum of "
                        f"{len(valid_args)} positional arguments but "
                        f"{len(args)} were given"
                    )
                if kwarg in kwargs:
                    raise TypeError(
                        f"{method.__name__}() got multiple values for "
                        f"argument '{kwarg}'"
                    )
                else:
                    kwargs[kwarg] = arg

            mapped_kwargs = dict()
            for key, value in kwargs.items():
                try:
                    mapped_key = valid_args[key]
                except KeyError:
                    raise TypeError(
                        f"{method.__name__}() got an unexpected keyword "
                        f"argument '{key}'"
                    )
                if isinstance(mapped_key, list):
                    for mkey in mapped_key:
                        if mkey not in mapped_kwargs:
                            mapped_kwargs[mkey] = value
                else:
                    mapped_kwargs[mapped_key] = value

            if style is not None:
                mapped_kwargs = {
                    **styles.get(style, method.__name__),
                    **mapped_kwargs,
                }

            if conditions is not None:
                for key, value in conditions.items():
                    if isinstance(value, dict):
                        if key in mapped_kwargs:
                            mapped_kwargs = {**value, **mapped_kwargs}
                    else:
                        mapped_kwargs = {**{key: value}, **mapped_kwargs}
            self.register(magics_macro(z_index=z_index, **mapped_kwargs))

        return wrapper

    return decorator



# (C) Copyright 2021- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from Magics import macro as magics


class Macro:

    PLOTTER = False

    def __init__(self, z_index=1, **kwargs):
        self.z_index = z_index
        self._kwargs = kwargs

    @property
    def macro(self):
        return self.__class__.__name__

    @property
    def kwargs(self):
        return {k: self._sanitise_value(v) for k, v in self._kwargs.items()}

    def execute(self):
        macro = getattr(magics, self.macro)
        return macro(**self.kwargs)

    @staticmethod
    def _sanitise_value(value):
        if isinstance(value, bool):
            value = bool_to_string(value)
        return value


def bool_to_string(boolean):
    return "on" if boolean else "off"


class output(Macro):
    pass


class mtext(Macro):
    pass


class mcont(Macro):
    PLOTTER = True


class mwind(Macro):
    pass


class mlegend(Macro):
    pass


class mmap(Macro):
    pass


class page(Macro):
    pass


class mcoast(Macro):
    pass


class mgrib(Macro):
    pass


class mnetcdf(Macro):
    pass



# (C) Copyright 2021- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os

import Magics.macro as magics

from . import data, macro
from .action import action

ARROW_STYLES = ["angle", "triangle", "triangle2", "triangle3"]


class GeoMap:
    """
    Class for designing and plotting geospatial maps.

    Parameters
    ----------
    area_name : str, optional
        The name of a an area (e.g. 'europe') with a pre-configured projection
        and extent - see examples for a sample of valid area names.
    projection : str, optional
        The name of the map projection to use for this map. See examples for
        a sample of valid projections.
    extent : tuple, optional
        A four-element list/tuple containing the latitude and longitude
        extents to use in the map. These must be provided in the order: lower-
        left latitude, lower-left longitude, upper-right latitude, upper-right
        longitude.
    """

    def __init__(self, *args, **kwargs):
        self._sources = []

        self.queue = []
        self._map(*args, **kwargs)

        self._show_legend = False

    def register(self, item):
        if item.__class__.__name__ == "page":
            self.queue = [self.queue[0]] + [item] + self.queue[1:]
        else:
            self.queue.append(item)

    def _map(self, *args, **kwargs):
        extent = kwargs.pop("extent", None)
        if extent is not None:
            (
                kwargs["lower_left_lat"],
                kwargs["lower_left_lon"],
                kwargs["upper_right_lat"],
                kwargs["upper_right_lon"],
            ) = extent
        self.__map(*args, **kwargs)

    @action(
        macro.mmap,
        {
            "page_id_line": False,
            "subpage_map_area_name": {"subpage_map_library_area": True},
            "subpage_expand_mode": True,
            "subpage_clipping": True,
        },
        area_name="subpage_map_area_name",
        projection="subpage_map_projection",
        lower_left_lat="subpage_lower_left_latitude",
        lower_left_lon="subpage_lower_left_longitude",
        upper_right_lat="subpage_upper_right_latitude",
        upper_right_lon="subpage_upper_right_longitude",
    )
    def __map(self, *args, **kwargs):
        pass

    @action(
        macro.page,
        {
            "page_id_line": True,
            "page_id_line_logo_name": {"page_id_line_logo_plot": True},
            "page_id_line_logo_plot": False,
            "page_id_line_system_plot": False,
            "page_id_line_date_plot": False,
            "page_id_line_magics": False,
            "page_id_line_errors_plot": False,
            "page_id_line_colour": "charcoal",
        },
        text="page_id_line_user_text",
        font="page_id_line_font",
        font_style="page_id_line_font_style",
        font_size="page_id_line_height",
        font_colour="page_id_line_colour",
        logo="page_id_line_logo_name",
        datestamp="page_id_line_date_plot",
    )
    def footer(self, *args, **kwargs):
        """
        Add a footer to the bottom of the plot, containing text and/or a logo.

        Parameters
        ----------
        text : str, optional
            A string of text to include in the footer.
        font : str, default='sansserif'
            The name of the font to use for footer text.
        font_style : str, default='normal'
            Style options for the footer font, e.g. `'bold'`.
        font_size : float, optional
            The font size to use, in cm.
        font_colour : str, default='charcoal'
            Either a hexadecimal colour or a named colour to use for the
            footer font.
        logo : bool or str, optional
            The name of an organisation or project whose logo should be added
            to the footer. Must be one of `'ecmwf'`, `'c3s'` or `'cams'` - or
            `False` if no logo should be included (default).
        datestamp : bool, optional
            If `True`, the date and time at which the map was generated will
            be included in the footer text.
        """
        pass

    @action(
        macro.mcoast,
        {
            "map_rivers": True,
            "map_coastline": True,
            "map_coastline_thickness": 0,
            "map_grid": False,
            "map_label": False,
        },
        resolution="map_coastline_resolution",
        line_colour="map_rivers_colour",
        line_style="map_rivers_style",
        line_thickness="map_rivers_thickness",
    )
    def rivers(self, *args, **kwargs):
        """
        Add rivers to the map.

        Parameters
        ----------
        resolution : str, default='low'
            The resolution of the rivers to be included in the map. Must be
            one of `'low'`, `'medium'` or `'high'`.
        line_colour : str, default='blue'
            Either a hexadecimal colour or a named colour to use for the
            river lines.
        line_style : str, default='solid'
            One of `'solid'`, `'dash'`, `'dot'`, `'chain_dash'`, or
            `'chain_dot'`.
        line_thickness : float, default=1.0
            The thickness of the river lines.
        """
        pass

    @action(
        macro.mcoast,
        {
            "map_coastline": True,
            "map_grid": False,
            "map_label": False,
            "map_coastline_land_shade_colour": {
                "map_coastline_land_shade": True,
            },
            "map_coastline_sea_shade_colour": {
                "map_coastline_sea_shade": True,
            },
        },
        resolution="map_coastline_resolution",
        line_colour="map_coastline_colour",
        line_style="map_coastline_style",
        line_thickness="map_coastline_thickness",
        land_colour="map_coastline_land_shade_colour",
        ocean_colour="map_coastline_sea_shade_colour",
    )
    def coastlines(self, *args, **kwargs):
        """
        Add coastlines to the map.

        Parameters
        ----------
        resolution : str, default='low'
            The resolution of the coastlines to be included in the map. Must
            be one of `'low'`, `'medium'` or `'high'`.
        line_colour : str, default='black'
            Either a hexadecimal colour or a named colour to use for the
            coastlines.
        line_style : str, default='solid'
            One of `'solid'`, `'dash'`, `'dot'`, `'chain_dash'`, or
            `'chain_dot'`.
        line_thickness : float, default=1.0
            The thickness of the coastlines.
        land_colour : str, optional
            Either a hexadecimal colour or a named colour to use for the fill
            colour of areas within coastline polygons (i.e. land).
        ocean_colour : str, optional
            Either a hexadecimal colour or a named colour to use for the fill
            colour of areas outside coastlines (i.e. oceans).
        """
        pass

    @action(
        macro.mcoast,
        {
            "map_coastline": False,
            "map_grid": True,
        },
        lat_frequency="map_grid_latitude_increment",
        lon_frequency="map_grid_longitude_increment",
        lat_reference="map_grid_latitude_reference",
        lon_reference="map_grid_longitude_reference",
        line_colour="map_grid_colour",
        line_style="map_grid_line_style",
        line_thickness="map_grid_thickness",
        labels="map_label",
        label_font="map_label_font_style",
        label_font_size="map_label_height",
        label_font_colour="map_label_colour",
        label_latitude_frequency="map_label_latitude_frequency",
        label_longitude_frequency="map_label_longitude_frequency",
        label_top_edge="map_label_top",
        label_bottom_edge="map_label_bottom",
        label_left_edge="map_label_left",
        label_right_edge="map_label_right",
    )
    def gridlines(self, *args, **kwargs):
        """
        Add gridlines to the map.

        Parameters
        ----------
        lat_frequency : float, default=10
            The interval in degrees of latitude between each latitude grid
            line.
        lon_frequency : float, default=20
            The interval in degrees of longitude between each longitude grid
            line.
        lat_reference : float, default=0
            The reference/starting latitude from which to begin drawing
            latitude lines at a frequency given by `lat_frequency`.
        lon_reference : float, default=0
            The reference/starting longitude from which to begin drawing
            longitude lines at a frequency given by `lon_frequency`.
        line_colour : str, default='blue'
            Either a hexadecimal colour or a named colour to use for the
            river lines.
        line_style : str, default='solid'
            One of `'solid'`, `'dash'`, `'dot'`, `'chain_dash'`, or
            `'chain_dot'`.
        line_thickness : float, default=1.0
            The thickness of the river lines.
        labels : bool, optional
            If `True`, gridlines will be given latitude and longitude labels.
        label_font : str, default='sansserif'
            The name of the font to use for gridline label text.
        label_font_size : float, optional
            The font size to use, in cm.
        label_font_colour : str, default='charcoal'
            Either a hexadecimal colour or a named colour to use for the
            gridline label font.
        label_latitude_frequency : int, optional
            The frequency at which to label latitude gridlines. A frequency of
            1 means every latitude gridline will be labelled.
        label_longitude_frequency : int, optional
            The frequency at which to label longitude gridlines. A frequency
            of 1 means every longitude gridline will be labelled.
        label_top_edge : bool, default=True
            If `True`, labels will be drawn where gridlines intersect the top
            edge/border of the map.
        label_bottom_edge : bool, default=True
            If `True`, labels will be drawn where gridlines intersect the
            bottom edge/border of the map.
        label_left_edge= : bool, default=True
            If `True`, labels will be drawn where gridlines intersect the
            left edge/border of the map.
        label_right_edge : bool, default=True
            If `True`, labels will be drawn where gridlines intersect the
            right edge/border of the map.
        """
        pass

    @action(
        macro.mtext,
        {
            "text_colour": "charcoal",
        },
        text="text_lines",
        text_colour="text_colour",
    )
    def title(self, text, **kwargs):
        """
        Add a title above the map.
        """
        pass

    def _input(self, source, **kwargs):
        source = data.detect_source(source, self)
        self._sources.append(source)
        source.get(**kwargs)

    def _vector_input(self, *args, wind_mode="uv", **kwargs):
        source = data.detect_vector_source(*args, wind_mode=wind_mode, geomap=self)
        self._sources.append(source)
        source.get(**kwargs)

    def contour_lines(self, source, *args, style=None, **kwargs):
        """
        Plot line contours on a map.
        """
        self._input(source)
        self._contour_lines(*args, style=style, **kwargs)

    def contour_shaded(self, source, *args, style=None, **kwargs):
        """
        Plot filled contours on a map.
        """
        self._input(source)
        self._contour_shaded(*args, style=style, **kwargs)

    def arrows(
        self,
        *,
        u=None,
        v=None,
        speed=None,
        direction=None,
        shaded=False,
        style=None,
        **kwargs,
    ):
        """
        Plot arrows on a map.
        """
        if all((u, v)):
            self._vector_input(u, v, wind_mode="uv")
        elif all((speed, direction)):
            self._vector_input(speed, direction, wind_mode="sd")
        else:
            raise TypeError("arrows() requires u and v OR speed and direction")

        method = self._arrows if not shaded else self._arrows_shaded

        return self._vector(method, style=style, **kwargs)

    def _vector(self, plotter, *args, **kwargs):

        arrow_head = kwargs.pop("arrow_head", None)
        if arrow_head is not None:
            if "-" in arrow_head:
                style, angle = arrow_head.split("-")
            else:
                style, angle = arrow_head, 45
            kwargs["_arrow_shape"] = ARROW_STYLES.index(style)
            kwargs["_arrow_ratio"] = int(angle) / 90

        plotter(*args, **kwargs)

    @action(
        macro.mcont,
        {
            "legend": True,
            "contour": False,
            "contour_shade": True,
            "contour_shade_method": "area_fill",
            "contour_interval": {
                "contour_level_selection_type": "interval",
            },
            "contour_level_list": {
                "contour_level_selection_type": "level_list",
            },
            "contour_label_text": {
                "contour_label_type": "text",
            },
            "contour_shade_palette_name": {
                "contour_shade_colour_method": "palette",
            },
        },
        levels="contour_level_list",
        interval="contour_interval",
        interval_reference="contour_reference_level",
        dynamic_levels="contour_level_count",
        fill_pattern="contour_shade_method",
        dot_size="contour_shade_dot_size",
        dot_max_density="contour_shade_max_level_density",
        dot_min_density="contour_shade_min_level_density",
        hatch_index="contour_shade_hatch_index",
        hatch_thickness="contour_shade_hatch_thickness",
        hatch_density="contour_shade_hatch_density",
        shade_type="contour_shade_technique",
        contour_method="contour_method",
        legend="legend",
    )
    def _contour_shaded(self, *args, **kwargs):
        pass

    @action(
        macro.mcont,
        {
            "legend": False,
            "contour": True,
            "contour_shade": False,
            "contour_interval": {
                "contour_level_selection_type": "interval",
            },
            "contour_level_list": {
                "contour_level_selection_type": "level_list",
            },
            "contour_label_text": {
                "contour_label_type": "text",
            },
        },
        line_style=["contour_line_style", "contour_highlight_style"],
        line_colour=["contour_line_colour", "contour_highlight_colour"],
        line_thickness=["contour_line_thickness", "contour_highlight_thickness"],
        highlight="contour_highlight",
        highlight_colour="contour_highlight_colour",
        highlight_style="contour_highlight_style",
        highlight_thickness="contour_highlight_thickness",
        highlight_frequency="contour_highlight_frequency",
        levels="contour_level_list",
        interval="contour_interval",
        interval_reference="contour_reference_level",
        dynamic_levels="contour_level_count",
        labels="contour_label",
        label_text="contour_label_text",
        label_size="contour_label_height",
        label_blanks="contour_label_blanking",
        label_quality="contour_label_quality",
        label_font="contour_label_font",
        label_font_style="contour_label_font_style",
        label_colour="contour_label_colour",
        label_frequency="contour_label_frequency",
        legend="legend",
    )
    def _contour_lines(self, *args, **kwargs):
        pass

    @action(
        macro.mwind,
        {
            "legend": False,
            "wind_thinning_method": "automatic",
            "wind_arrow_calm_below": {"wind_arrow_calm_indicator": True},
            "wind_flag_calm_below": {"wind_flag_calm_indicator": True},
        },
        wind_style="wind_field_type",
        colour=["wind_flag_colour", "wind_arrow_colour"],
        flag_length="wind_flag_length",
        flag_origin_marker="wind_flag_origin_marker",
        flag_origin_size="wind_flag_origin_marker_size",
        density="wind_thinning_factor",
        calm_threshold=["wind_arrow_calm_below", "wind_flag_calm_below"],
        calm_indicator_size=[
            "wind_arrow_calm_indicator_size",
            "wind_flag_calm_indicator_size",
        ],
        _arrow_shape="wind_arrow_head_shape",
        _arrow_ratio="wind_arrow_head_ratio",
        max_speed=["wind_arrow_max_speed", "wind_flag_max_speed"],
        min_speed=["wind_arrow_min_speed", "wind_flag_min_speed"],
        arrow_origin="wind_arrow_origin_position",
        line_thickness=["wind_arrow_thickness", "wind_flag_thickness"],
        line_style=["wind_arrow_style", "wind_flag_style"],
        legend="legend",
    )
    def _arrows(self, *args, **kwargs):
        pass

    @action(
        macro.mwind,
        {
            "legend": False,
            "wind_advanced_method": True,
            "wind_thinning_method": "automatic",
            "wind_arrow_calm_below": {"wind_arrow_calm_indicator": True},
            "wind_flag_calm_below": {"wind_flag_calm_indicator": True},
            "contour_level_count": {
                "wind_advanced_colour_selection_type": "count",
            },
            "contour_interval": {
                "wind_advanced_colour_selection_type": "interval",
            },
            "contour_level_list": {
                "wind_advanced_colour_selection_type": "list",
            },
            "wind_advanced_colour_list": {
                "wind_advanced_colour_table_colour_method": "list",
            },
            "contour_shade_min_level_colour": {
                "wind_advanced_colour_table_colour_method": "calculate",
            },
        },
        wind_style="wind_field_type",
        flag_length="wind_flag_length",
        flag_origin_marker="wind_flag_origin_marker",
        flag_origin_size="wind_flag_origin_marker_size",
        density="wind_thinning_factor",
        calm_threshold=["wind_arrow_calm_below", "wind_flag_calm_below"],
        calm_indicator_size=[
            "wind_arrow_calm_indicator_size",
            "wind_flag_calm_indicator_size",
        ],
        _arrow_shape="wind_arrow_head_shape",
        _arrow_ratio="wind_arrow_head_ratio",
        max_speed=[
            "wind_arrow_max_speed",
            "wind_flag_max_speed",
            "contour_max_level",
            "contour_shade_max_level",
            "wind_advanced_colour_max_value",
        ],
        min_speed=[
            "wind_arrow_min_speed",
            "wind_flag_min_speed",
            "contour_min_level",
            "contour_shade_min_level",
            "wind_advanced_colour_min_value",
        ],
        arrow_origin="wind_arrow_origin_position",
        line_thickness=["wind_arrow_thickness", "wind_flag_thickness"],
        line_style=["wind_arrow_style", "wind_flag_style"],
        bin_count=["contour_level_count", "wind_advanced_colour_level_count"],
        bin_tolerance=[
            "contour_level_tolerance",
            "wind_advanced_colour_level_tolerance",
        ],
        bin_reference=[
            "contour_reference_level",
            "wind_advanced_colour_reference_level",
        ],
        bin_interval=["contour_interval", "wind_advanced_colour_level_interval"],
        bins=["contour_level_list", "wind_advanced_colour_level_list"],
        colours="wind_advanced_colour_list",
        min_colour=[
            "contour_shade_min_level_colour",
            "wind_advanced_colour_min_level_colour",
        ],
        max_colour=[
            "contour_shade_max_level_colour",
            "wind_advanced_colour_max_level_colour",
        ],
        colour_wheel_direction=[
            "contour_shade_colour_direction",
            "wind_advanced_colour_direction",
        ],
        legend="legend",
    )
    def _arrows_shaded(self, *args, **kwargs):
        pass

    def legend(self, *args, **kwargs):
        self._show_legend = True
        position = kwargs.pop("position", None)
        if position is not None:
            if isinstance(position, (list, tuple)):
                try:
                    (
                        kwargs["_user_x_position"],
                        kwargs["_user_y_position"],
                        kwargs["_user_x_length"],
                        kwargs["_user_y_length"],
                    ) = position
                except ValueError:
                    raise ValueError("position expects [x, x_len, y, y_len]")
            else:
                kwargs["_auto_position"] = position
        return self._legend(*args, **kwargs)

    @action(
        macro.mlegend,
        {
            "legend_display_type": "continuous",
            "legend_text_colour": "charcoal",
            "legend_title_font_colour": "charcoal",
            "legend_title_text": {"legend_title": True},
            "legend_user_minimum_text": {"legend_user_minimum": True},
            "legend_user_maximum_text": {"legend_user_maximum": True},
            "legend_box_x_position": {"legend_box_mode": "positional"},
        },
        text_colour=["legend_text_colour", "legend_title_font_colour"],
        title="legend_title_text",
        title_text_colour="legend_title_font_colour",
        title_orientation="legend_title_orientation",
        title_font_size="legend_title_font_size",
        title_position="legend_title_position",
        minimum_text="legend_user_minimum_text",
        maximum_text="legend_user_maximum_text",
        display_type="legend_display_type",
        label_frequency="legend_label_frequency",
        _auto_position="legend_automatic_position",
        _user_x_position="legend_box_x_position",
        _user_y_position="legend_box_y_position",
        _user_x_length="legend_box_x_length",
        _user_y_length="legend_box_y_length",
    )
    def _legend(self, *args, **kwargs):
        pass

    def show(self):
        return self._execute()

    def save(self, *args, **kwargs):
        nargs = len(args)
        if nargs == 1:
            name, format = os.path.splitext(args[0])
            kwargs["output_name"] = name
            kwargs["output_name_first_page_number"] = False
            kwargs["output_formats"] = [format.lstrip(".")]
        elif nargs > 1:
            raise TypeError(f"save expected at most 1 argument, got {nargs}")
        output = macro.output(**kwargs)
        return self._execute(output=output)

    def _execute(self, output=None):
        if not self._show_legend:
            for item in self.queue:
                if item.PLOTTER:
                    item._kwargs["legend"] = False
        output = [output] if output is not None else []
        queue = output + sorted(self.queue, key=lambda x: x.z_index)
        result = magics.plot(*(macro.execute() for macro in queue))

        self._sources = []

        return result




class Data:

    def __init__(self, source, geomap, *args, **kwargs):
        self.source = source
        self.geomap = geomap
        self.args = args
        self.kwargs = kwargs

    def register(self, *args, **kwargs):
        self.geomap.register(*args, **kwargs)

    def get(self):
        return self._get(self.source, *self.args, **self.kwargs)

    def _get(self, *args, **kwargs):
        raise NotImplementedError

    def get_extent(self):
        return dict()



from ._helpers import get_active_geomap


def quick_plot(func):
    def plotter(*args, **kwargs):
        geomap = get_active_geomap()
        getattr(geomap, func.__name__)(*args, **kwargs)
        return geomap

    return plotter


def show():
    return get_active_geomap().show()


@quick_plot
def contour_lines(*args, **kwargs):
    pass


@quick_plot
def wind(*args, **kwargs):
    pass


@quick_plot
def rivers(*args, **kwargs):
    pass




import os

from . import ecdata, grib, netcdf, temporary


def detect_source(source, geomap):
    if isinstance(source, str):
        data = detect_file(source)(source, geomap)
    else:
        data = ecdata.EcData(source, geomap)
    return data


def detect_vector_source(*args, wind_mode, geomap):
    tmp = temporary.temp_file('.grib').path
    indices = []
    for source in args:
        if not isinstance(source, str):
            fname, grib_index = source.grib_index()[0]
        else:
            raise NotImplementedError("vector plots only work with ecmwf-data")
        with open(tmp, 'ab') as tf:
            with open(fname, 'rb') as f:
                for line in f:
                    tf.write(line)
        indices.append(grib_index)
    return grib.Grib(tmp, geomap, wind_mode=wind_mode, grib_mode="byte_offset", wind_1=indices[0], wind_2=indices[1])


def detect_file(file_name):
    _, ext = os.path.splitext(file_name)
    ext = ext.lstrip(".")
    format_types = {
        "grib": "grib",
        "grb": "grib",
        "grib1": "grib",
        "grib2": "grib",
        "nc": "netcdf",
        "nc3": "netcdf",
        "nc4": "netcdf",
        "cdf": "netcdf",
    }
    try:
        format_type = format_types[ext]
    except KeyError:
        f"unrecognised extension '.{ext}'; try grib or netcdf instead"

    return {
        "grib": grib.Grib,
        "netcdf": netcdf.NetCDF,
    }[format_type]




from .grib import Grib


class EcData(Grib):

    def get(self, **kwargs):
        file_name, index = self.source.grib_index()[0]
        return self._get(
            file_name, index,
            grib_mode="byte_offset",
            *self.args, **{**self.kwargs, **kwargs},
        )
